@@ -9,11 +9,13 @@ import (
 	"os"
 	"path"
 	"regexp"
+	"strconv"
 	"strings"
 	"sync"
 	"time"
 
 	"github.com/spf13/cobra"
+	"k8s.io/klog/v2"
 
 	corev1 "k8s.io/api/core/v1"
 	rbacv1 "k8s.io/api/rbac/v1"
@@ -33,8 +35,10 @@ import (
 
 	imagev1client "github.com/openshift/client-go/image/clientset/versioned/typed/image/v1"
 	"github.com/openshift/library-go/pkg/image/imageutil"
+	imagereference "github.com/openshift/library-go/pkg/image/reference"
 	"github.com/openshift/library-go/pkg/operator/resource/retry"
 
+	"github.com/openshift/oc/pkg/cli/admin/inspect"
 	"github.com/openshift/oc/pkg/cli/rsync"
 )
 
@@ -55,6 +59,9 @@ var (
 		# gather information with a specific local folder to copy to
 		  oc adm must-gather --dest-dir=/local/directory
 
+		# gather audit information
+		  oc adm must-gather -- /usr/bin/gather_audit_logs
+
 		# gather information using multiple plug-in images
 		  oc adm must-gather --image=quay.io/kubevirt/must-gather --image=quay.io/openshift/origin-must-gather
 
@@ -85,7 +92,7 @@ func NewMustGatherCommand(f kcmdutil.Factory, streams genericclioptions.IOStream
 	cmd.Flags().StringSliceVar(&o.ImageStreams, "image-stream", o.ImageStreams, "Specify an image stream (namespace/name:tag) containing a must-gather plugin image to run.")
 	cmd.Flags().StringVar(&o.DestDir, "dest-dir", o.DestDir, "Set a specific directory on the local machine to write gathered data to.")
 	cmd.Flags().StringVar(&o.SourceDir, "source-dir", o.SourceDir, "Set the specific directory on the pod copy the gathered data from.")
-	cmd.Flags().Int64Var(&o.Timeout, "timeout", 600, "The length of time to gather data, in seconds. Defaults to 10 minutes.")
+	cmd.Flags().StringVar(&o.timeoutStr, "timeout", "10m", "The length of time to gather data, like 5s, 2m, or 3h, higher than zero. Defaults to 10 minutes.")
 	cmd.Flags().BoolVar(&o.Keep, "keep", o.Keep, "Do not delete temporary resources when command completes.")
 	cmd.Flags().MarkHidden("keep")
 
@@ -97,6 +104,7 @@ func NewMustGatherOptions(streams genericclioptions.IOStreams) *MustGatherOption
 		SourceDir: "/must-gather/",
 		IOStreams: streams,
 		LogOut:    newPrefixWriter(streams.Out, "[must-gather      ] OUT"),
+		Timeout:   10 * time.Minute,
 	}
 }
 
@@ -117,6 +125,21 @@ func (o *MustGatherOptions) Complete(f kcmdutil.Factory, cmd *cobra.Command, arg
 	} else {
 		o.Command = args
 	}
+	if len(o.timeoutStr) > 0 {
+		if strings.ContainsAny(o.timeoutStr, "smh") {
+			o.Timeout, err = time.ParseDuration(o.timeoutStr)
+			if err != nil {
+				return fmt.Errorf(`invalid argument %q for "--timeout" flag: %v`, o.timeoutStr, err)
+			}
+		} else {
+			fmt.Fprint(o.ErrOut, "Flag --timeout's value in seconds has been deprecated, use duration like 5s, 2m, or 3h, instead\n")
+			i, err := strconv.ParseInt(o.timeoutStr, 0, 64)
+			if err != nil {
+				return fmt.Errorf(`invalid argument %q for "--timeout" flag: %v`, o.timeoutStr, err)
+			}
+			o.Timeout = time.Duration(i) * time.Second
+		}
+	}
 	if len(o.DestDir) == 0 {
 		o.DestDir = fmt.Sprintf("must-gather.local.%06d", rand.Int63())
 	}
@@ -205,7 +228,8 @@ type MustGatherOptions struct {
 	Images       []string
 	ImageStreams []string
 	Command      []string
-	Timeout      int64
+	Timeout      time.Duration
+	timeoutStr   string
 	Keep         bool
 
 	RsyncRshCmd string
@@ -234,7 +258,8 @@ func (o *MustGatherOptions) Run() error {
 				"openshift.io/run-level": "0",
 			},
 			Annotations: map[string]string{
-				"oc.openshift.io/command": "oc adm must-gather",
+				"oc.openshift.io/command":    "oc adm must-gather",
+				"openshift.io/node-selector": "",
 			},
 		},
 	}, metav1.CreateOptions{})
@@ -271,6 +296,12 @@ func (o *MustGatherOptions) Run() error {
 	// ... and finally must-gather pod
 	var pods []*corev1.Pod
 	for _, image := range o.Images {
+		_, err := imagereference.Parse(image)
+		if err != nil {
+			o.log("unable to parse image reference %s: %v", image, err)
+			return err
+		}
+
 		pod, err := o.Client.CoreV1().Pods(ns.Name).Create(context.TODO(), o.newPod(o.NodeName, image), metav1.CreateOptions{})
 		if err != nil {
 			return err
@@ -290,7 +321,7 @@ func (o *MustGatherOptions) Run() error {
 
 	var wg sync.WaitGroup
 	wg.Add(len(pods))
-	errs := make(chan error, len(pods))
+	errCh := make(chan error, len(pods))
 	for _, pod := range pods {
 		go func(pod *corev1.Pod) {
 			defer wg.Done()
@@ -300,7 +331,7 @@ func (o *MustGatherOptions) Run() error {
 			// wait for gather container to be running (gather is running)
 			if err := o.waitForGatherContainerRunning(pod); err != nil {
 				log("gather did not start: %s", err)
-				errs <- fmt.Errorf("gather did not start for pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("gather did not start for pod %s: %s", pod.Name, err)
 				return
 			}
 			// stream gather container logs
@@ -312,7 +343,7 @@ func (o *MustGatherOptions) Run() error {
 			log("waiting for gather to complete")
 			if err := o.waitForGatherToComplete(pod); err != nil {
 				log("gather never finished: %v", err)
-				errs <- fmt.Errorf("gather never finished for pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("gather never finished for pod %s: %s", pod.Name, err)
 				return
 			}
 
@@ -321,23 +352,29 @@ func (o *MustGatherOptions) Run() error {
 			pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{})
 			if err != nil {
 				log("gather output not downloaded: %v\n", err)
-				errs <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
 				return
 			}
 			if err := o.copyFilesFromPod(pod); err != nil {
 				log("gather output not downloaded: %v\n", err)
-				errs <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
+				errCh <- fmt.Errorf("unable to download output from pod %s: %s", pod.Name, err)
 				return
 			}
 		}(pod)
 	}
 	wg.Wait()
-	close(errs)
-	var arr []error
-	for i := range errs {
-		arr = append(arr, i)
+	close(errCh)
+	var errs []error
+	for i := range errCh {
+		errs = append(errs, i)
 	}
-	return errors.NewAggregate(arr)
+
+	// now gather all the events into a single file and produce a unified file
+	if err := inspect.CreateEventFilterPage(o.DestDir); err != nil {
+		errs = append(errs, err)
+	}
+
+	return errors.NewAggregate(errs)
 }
 
 func newPodOutLogger(out io.Writer, podName string) func(string, ...interface{}) {
@@ -374,15 +411,23 @@ func (o *MustGatherOptions) copyFilesFromPod(pod *corev1.Pod) error {
 		Destination:   &rsync.PathSpec{PodName: "", Path: destDir},
 		Client:        o.Client,
 		Config:        o.Config,
+		Compress:      true,
 		RshCmd:        fmt.Sprintf("%s --namespace=%s -c copy", o.RsyncRshCmd, pod.Namespace),
 		IOStreams:     streams,
 	}
 	rsyncOptions.Strategy = rsync.NewDefaultCopyStrategy(rsyncOptions)
-	return rsyncOptions.RunRsync()
+	err := rsyncOptions.RunRsync()
+	if err != nil {
+		klog.V(4).Infof("re-trying rsync after initial failure %v", err)
+		// re-try copying data before letting it go
+		err = rsyncOptions.RunRsync()
+	}
+	return err
 }
 
 func (o *MustGatherOptions) getGatherContainerLogs(pod *corev1.Pod) error {
-	return (&logs.LogsOptions{
+	since2s := int64(2)
+	opts := &logs.LogsOptions{
 		Namespace:   pod.Namespace,
 		ResourceArg: pod.Name,
 		Options: &corev1.PodLogOptions{
@@ -395,7 +440,23 @@ func (o *MustGatherOptions) getGatherContainerLogs(pod *corev1.Pod) error {
 		ConsumeRequestFn: logs.DefaultConsumeRequest,
 		LogsForObject:    polymorphichelpers.LogsForObjectFn,
 		IOStreams:        genericclioptions.IOStreams{Out: newPrefixWriter(o.Out, fmt.Sprintf("[%s] POD", pod.Name))},
-	}).RunLogs()
+	}
+
+	for {
+		// gather script might take longer than the default API server time,
+		// so we should check if the gather script still runs and re-run logs
+		// thus we run this in a loop
+		if err := opts.RunLogs(); err != nil {
+			return err
+		}
+
+		// to ensure we don't print all of history set since to past 2 seconds
+		opts.Options.(*corev1.PodLogOptions).SinceSeconds = &since2s
+		if done, _ := o.isGatherDone(pod); done {
+			return nil
+		}
+		klog.V(4).Infof("lost logs, re-trying...")
+	}
 }
 
 func newPrefixWriter(out io.Writer, prefix string) io.Writer {
@@ -410,52 +471,55 @@ func newPrefixWriter(out io.Writer, prefix string) io.Writer {
 }
 
 func (o *MustGatherOptions) waitForGatherToComplete(pod *corev1.Pod) error {
-	err := wait.PollImmediate(time.Duration(o.Timeout)*time.Second, time.Minute, func() (bool, error) {
-		var err error
-		if pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{}); err != nil {
-			// at this stage pod should exist, we've been gathering container logs, so error if not found
-			if kerrors.IsNotFound(err) {
-				return true, err
-			}
-			return false, nil
-		}
-		var state *corev1.ContainerState
-		for _, cstate := range pod.Status.ContainerStatuses {
-			if cstate.Name == "gather" {
-				state = &cstate.State
-				break
-			}
-		}
+	return wait.PollImmediate(10*time.Second, o.Timeout, func() (bool, error) {
+		return o.isGatherDone(pod)
+	})
+}
 
-		// missing status for gather container => timeout in the worst case
-		if state == nil {
-			return false, nil
-		}
-
-		if state.Terminated != nil {
-			if state.Terminated.ExitCode == 0 {
-				return true, nil
-			}
-			return true, fmt.Errorf("%s/%s unexpectedly terminated: exit code: %v, reason: %s, message: %s", pod.Namespace, pod.Name, state.Terminated.ExitCode, state.Terminated.Reason, state.Terminated.Message)
+func (o *MustGatherOptions) isGatherDone(pod *corev1.Pod) (bool, error) {
+	var err error
+	if pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{}); err != nil {
+		// at this stage pod should exist, we've been gathering container logs, so error if not found
+		if kerrors.IsNotFound(err) {
+			return true, err
 		}
 		return false, nil
-	})
-	if err != nil {
-		return err
 	}
-	return nil
+	var state *corev1.ContainerState
+	for _, cstate := range pod.Status.ContainerStatuses {
+		if cstate.Name == "gather" {
+			state = &cstate.State
+			break
+		}
+	}
+
+	// missing status for gather container => timeout in the worst case
+	if state == nil {
+		return false, nil
+	}
+
+	if state.Terminated != nil {
+		if state.Terminated.ExitCode == 0 {
+			return true, nil
+		}
+		return true, fmt.Errorf("%s/%s unexpectedly terminated: exit code: %v, reason: %s, message: %s", pod.Namespace, pod.Name, state.Terminated.ExitCode, state.Terminated.Reason, state.Terminated.Message)
+	}
+	return false, nil
 }
 
 func (o *MustGatherOptions) waitForGatherContainerRunning(pod *corev1.Pod) error {
-	return wait.PollImmediate(time.Duration(o.Timeout)*time.Second, time.Minute, func() (bool, error) {
+	return wait.PollImmediate(10*time.Second, o.Timeout, func() (bool, error) {
 		var err error
 		if pod, err = o.Client.CoreV1().Pods(pod.Namespace).Get(context.TODO(), pod.Name, metav1.GetOptions{}); err == nil {
 			if len(pod.Status.ContainerStatuses) == 0 {
 				return false, nil
 			}
 			state := pod.Status.ContainerStatuses[0].State
-			if state.Waiting != nil && state.Waiting.Reason == "ErrImagePull" {
-				return true, fmt.Errorf("unable to pull image: %v: %v", state.Waiting.Reason, state.Waiting.Message)
+			if state.Waiting != nil {
+				switch state.Waiting.Reason {
+				case "ErrImagePull", "ImagePullBackOff", "InvalidImageName":
+					return true, fmt.Errorf("unable to pull image: %v: %v", state.Waiting.Reason, state.Waiting.Message)
+				}
 			}
 			running := state.Running != nil
 			terminated := state.Terminated != nil
@@ -496,6 +560,14 @@ func (o *MustGatherOptions) newClusterRoleBinding(ns string) *rbacv1.ClusterRole
 // - copy: no-op container we can exec into
 func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 	zero := int64(0)
+
+	nodeSelector := map[string]string{
+		corev1.LabelOSStable: "linux",
+	}
+	if node == "" {
+		nodeSelector["node-role.kubernetes.io/master"] = ""
+	}
+
 	ret := &corev1.Pod{
 		ObjectMeta: metav1.ObjectMeta{
 			GenerateName: "must-gather-",
@@ -516,9 +588,11 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 			},
 			Containers: []corev1.Container{
 				{
-					Name:    "gather",
-					Image:   image,
-					Command: []string{"/usr/bin/gather"},
+					Name:            "gather",
+					Image:           image,
+					ImagePullPolicy: corev1.PullIfNotPresent,
+					// always force disk flush to ensure that all data gathered is accessible in the copy container
+					Command: []string{"/bin/bash", "-c", "/usr/bin/gather; sync"},
 					VolumeMounts: []corev1.VolumeMount{
 						{
 							Name:      "must-gather-output",
@@ -528,9 +602,10 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 					},
 				},
 				{
-					Name:    "copy",
-					Image:   image,
-					Command: []string{"/bin/bash", "-c", "trap : TERM INT; sleep infinity & wait"},
+					Name:            "copy",
+					Image:           image,
+					ImagePullPolicy: corev1.PullIfNotPresent,
+					Command:         []string{"/bin/bash", "-c", "trap : TERM INT; sleep infinity & wait"},
 					VolumeMounts: []corev1.VolumeMount{
 						{
 							Name:      "must-gather-output",
@@ -540,6 +615,7 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 					},
 				},
 			},
+			NodeSelector:                  nodeSelector,
 			TerminationGracePeriodSeconds: &zero,
 			Tolerations: []corev1.Toleration{
 				{
@@ -549,7 +625,9 @@ func (o *MustGatherOptions) newPod(node, image string) *corev1.Pod {
 		},
 	}
 	if len(o.Command) > 0 {
-		ret.Spec.Containers[0].Command = o.Command
+		// always force disk flush to ensure that all data gathered is accessible in the copy container
+		ret.Spec.Containers[0].Command = []string{"/bin/bash", "-c", fmt.Sprintf("%s; sync", strings.Join(o.Command, " "))}
 	}
+
 	return ret
 }
